package Vgluefftw;     ## -*-CPerl-*-

use Exporter 'import';
require DynaLoader;
@ISA = qw(DynaLoader);
use Data::Dumper;

use bytes;
use warnings;
use strict;

=head1 NAME

Vgluefftw - Glue to FFTW for audio packets

=head1 VERSION

Version 0.04

=cut

our $VERSION = '0.02';

=head1 SYNOPSIS

Vgluefftw provides an FFTW-based routines that
looks for tones in audio packets.

    use Vgluefftw;

=head1 FUNCTIONS

=head2 ($tone_freq, $avg_mag, $max_mag, $threshhold, $hits) = fft_packet ($num_samples, $bytes_per_sample, $packet, $mag_threshhold, $moise_threshhold);

Takes a raw audio $packet containing $num_samples samples of
$bytes_per_sample, which must be either 1, 2, or 4.  If
$bytes_per_sample is 1, the input is assumed to be ulaw,
otherwise it is assumed to be signed linear in native ordering.

Also takes $mag_threshhold which is a floating point number > 1.0
representing the multiplier of the average magnitude to use
as a magnitude threshhold for determining a tone magnitude.

Also takes $noise_threshhold, an absolute tone magnitude
that is the smallest acceptable tone detection magnitude.

Returns:  ($tone_freq, $avg_mag, $max_mag, $threshhold, $hits, $data)
  $tone_freq = index of tone freq found (cycles per packet) or 0 if none
               (can be number halfway between two indexes)
  $avg_mag = average magnitude of tone freqs
  $max_mag = maximum magnitude of tone freqs
  $threshhold = absolute threshhold magnitude (computed from $mag_threshhold)
  $hits = number of fundamental tone frequencies that exceeded both
          the magnitude threshhold and the noise threshhold
  $data = raw data of FFT'd magnitudes

=cut

sub fft_packet
{
    my ($num_samples) = shift (@_);
    my ($bytes_per_sample) = shift (@_);
    my ($packet) = shift (@_);
    my ($mag_threshhold) = shift (@_);
    my ($noise_threshhold) = shift (@_);
    my ($freq, $avg, $max, $thresh, $hits, $data);

    ($freq, $avg, $max, $thresh, $hits) =
      Vgluefftw::_fft_packet
	  ($num_samples, $bytes_per_sample, $packet,
	   $mag_threshhold, $noise_threshhold);
    $data = unpack ("a" . $num_samples * $bytes_per_sample, $packet);
    return ($freq, $avg, $max, $thresh, $hits, $data);
}

=head2 ($ok, $msg, @$tones, @$raw_results) = find_tones ($file, $mag_thresh, $noise_thresh, $count_thresh, $show_details);

Given $file the name of a file containing raw ulaw audio,
returns @tones wich is a sequence of tones found in the file.
Each item in tones is a frequency, a multiple of 25Hz, with
about a 25Hz resolution.
Uses $mag_thresh and $noise_thresh as in fft_packet().
Uses $count_thresh as the number of consecutive packets that
must contain the frequency before it's considered present.
If $show_details is false, does not return @$raw_results.

@raw_results is a list of hashrefs, one for each 0.02 second sample:
  ->{"freq"}	-- frequency tone found, if any
  ->{"avg"}	-- average frequency magnitude
  ->{"max"}	-- maximum frequency magnitude
  ->{"thresh"}	-- computed frequency magnitude threshhold
  ->{"hits"}	-- how many frequencies exceeded threshhold
  ->{"found"}	-- if this packet marks a newly found tone
  ->{"mags"}	-- packet of FFT'd data, now magnitudes

If $ok is false, there was an error, and $msg is an error message.

=cut

sub find_tones
{
    my ($file) = shift (@_);
    my ($mag_thresh) = shift (@_);
    my ($noise_thresh) = shift (@_);
    my ($freq_index_count_threshhold) = shift (@_);
    my ($show_details) = shift (@_);
    my ($packetsize) = 160;
    my ($filesize, $num_packets, $i, $packet);
    my ($tone_freq_index, $avg_mag, $max_mag, $threshhold, $hits, $data);
    my ($last_freq_index) = 0;
    my ($cur_freq_index_count) = 0;
    my ($last_found_freq_index) = 0;
    my ($tones) = [];
    my ($raw_results) = [];
    my (@statinfo) = stat ($file);
    scalar (@statinfo) || return (0, "Cannot stat $file: $!");
    $filesize = $statinfo[7];
    $num_packets = int ($filesize / 160);
    ($num_packets > 0) || return (1, "");
    $last_freq_index = 0;
    open (Vgluefftw::FH, $file)
      || return (0, "Cannot open $file for reading: $!");
    for ($i = 0; $i < $num_packets; ++$i)
    {
	my ($packet_result) = {};

	##  Compute tone in next packet
	(read (Vgluefftw::FH, $packet, $packetsize) == $packetsize)
	  || return (0, "Error reading from $file: $!");
	($tone_freq_index, $avg_mag, $max_mag, $threshhold, $hits, $data) =
	  fft_packet ($packetsize, 1, $packet, $mag_thresh, $noise_thresh);
	if ($show_details)
	{
	    $packet_result =
	    {
	     "freq" => $tone_freq_index,
	     "avg" => $avg_mag,
	     "max" => $max_mag,
	     "thresh" => $threshhold,
	     "hits" => $hits,
	     "found" => 0,
	     "mags" => $data
	    };
	};

	##  De-bounce this tone (wait for it to stabilize)
	if (abs ($tone_freq_index - $last_freq_index) > .7)
	{
	    ##  Tone doesn't match previous freq
	    $last_freq_index = $tone_freq_index;
	    $cur_freq_index_count = 1;
	}
	else
	{
	    ##  Found same tone as previous packet
	    if (++$cur_freq_index_count == $freq_index_count_threshhold)
	    {
		##  Found a de-bounced tone
		if ($last_freq_index != 0)
		{
		    ##  Start of new frequency
		    push (@$tones, int ($tone_freq_index * 50));
		    if ($show_details)
		    {
			$packet_result->{"found"} = 1;
		    };
		}
	    };
	};

	if ($show_details)
	{
	    push (@$raw_results, $packet_result);
	};
    };
    close (Vgluefftw::FH);
    if ($show_details)
    {
	return (1, "", $tones, $raw_results);
    }
    else
    {
	return (1, "", $tones);
    };
}

=head1 AUTHOR

Doug Campbell, C<< <soup at ampersand.com> >>

=head1 BUGS

See the voiceglue.org web page for information on
reporting bugs.

=head1 SUPPORT

See the voiceglue.org web page for information on
support.

=head1 ACKNOWLEDGEMENTS

=head1 COPYRIGHT & LICENSE

Copyright 2006-2008 Ampersand Inc., Doug Campbell

This file is part of Vgluefftw.

Vgluefftw is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 3 of the License, or
(at your option) any later version.

Vgluefftw is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Vgluefftw; if not, see <http://www.gnu.org/licenses/>.

=cut

bootstrap Vgluefftw $VERSION;        ##  if using xs

1; # End of Vgluefftw
